const queryDashboardFiliais = require('../model/adqGetDashboardFiliais');
const getMaxMinRows = require('../helpers/getMaxMinRows');
const qryErroImportaVenda = require('../model/qryErroImportaVendaModel');

class DashboardController {
  async getUnidades(req, res) {
    try {
      const { page = 1, limit = 10 } = req.query;
      const { minRow, maxRow } = getMaxMinRows(page, limit);
      const queryResult = await queryDashboardFiliais.get({ MIN_ROW: minRow, MAX_ROW: maxRow });
      const { rows } = queryResult;

      if (!rows.length) {
        return res.status(400).json({ offices: [] });
      }

      const offices = rows.map((item) => {
        return {
          cdFilial: item[0],
          nmFilial: item[1],
          idStatusConexaoSync: 'I',
        };
      });

      return res.status(200).json({ offices });
    } catch (error) {
      return res.status(500).json({ error });
    }
  }

  async getErroImportaVenda(req, res) {
    try {
      const { page = 1, limit = 10 } = req.query;
      const { minRow, maxRow } = getMaxMinRows(page, limit);
      const queryResult = await qryErroImportaVenda.get({ MIN_ROW: minRow, MAX_ROW: maxRow });
      const { rows } = queryResult;

      if (!rows.length) {
        return res.status(400).json({ errosImporta: [] });
      }

      const errosImporta = rows.map((item) => {
        return {
          id: item[0],
          chave: item[1],
          tabela: item[2],
          dsinstrucao: item[3],
          dsparams: item[4],
          dserro: item[5],
          dterro: item[6],
        };
      });

      return res.status(200).json({ errosImporta });
    } catch (error) {
      return res.status(500).json({ error });
    }
  }
}

module.exports = new DashboardController();
