const io = require('socket.io');
const AuthService = require('./AuthService');
const UpService = require('./UpService');
const adqGetDashboardFiliais = require('../model/adqGetDashboardFiliais');
const parametrizacoesModel = require('../model/parametrizacoes');
const intervalMonitoring = require('../functions/IntervalMonitoring');
const paginacao = require('../functions/Paginacao');
const updaterController = require('../controller/UpdaterController');
const monitoringController = require('../controller/MonitoringController');

class SocketService {
  async start(http) {
    this.clientlist = {};
    this.browserlist = {};
    this.slavelist = await this.getDashboardFiliais();
    this.defaultInitTimeout = 5000;
    this.intervals = []; // Registrar is intervals
    this.conexoes = []; // Registrar os portais
    this.assIntervals = [];
    this.lastSocketId = null;

    this.parametersError = [];
    this.vendasError = [];
    this.satError = [];
    this.monitoringResult = {};
    this.io = io(http, { cors: '*' });

    this.io.on('connection', (socket) => {
      this.socket = socket;
      console.log('newconnection: ', socket.id);
      socket.emit('updateVersion', { serviceName: 'xxx' });

      /**
       * Atualiza informações do Dashboard, inclusive o ID do socket e registra a sala
       */
      socket.on('join', (room) => {
        this.slavelist.filter((filial) => {
          if(filial.id === room.id){
            Object.assign(filial, room);
          }
          return filial;
        });
        socket.join(room.room);
      });

      /** Garante que o portal está conectado */
      socket.on('ping', (data) => {
        const clintId = (data.id)? data.id : data;
        if(data.filial) {
          /** se for loja, atualiza os dados e avise o portal */
          this.slavelist.filter((filial) => {
            if(filial.id === data.filial) {
              /** Merge de info */
              Object.assign(filial, {socketId: data.id, situacao: 'Ativo'});
              /** Atualiza o Portal */
              const portalList = Object.keys(this.clientlist).map((item) => {
                return this.clientlist[item].id;
              });
              this.io.to(portalList).emit('login', {
                text: 'UNIDADE CONECTADA',
                error: false,
                message: {
                  data: this.slavelist,
                },
              });
            }
            return filial;
          });
        }

        /** Confere as conexões */
        this.conexoes.filter((portais) => {
          if(portais.id === clintId){
            Object.assign(portais, {time: new Date().getTime()});
          }
          return portais;
        });
      });

      /** Remove conexões anativas */
      this.intervals.push({
        id: socket.id,
        interval: setInterval(() => {
          console.log('Conectado: ', socket.id, this.conexoes);
          this.conexoes.forEach((portais) => {
            if((portais.time + 1200000) < new Date().getTime()){
              /** Client não atualizou a conexão, mateo */
              // socket.disconnect();
              this.io.sockets.sockets.forEach((socket) => {
                // console.log('Deveria desconectar o: ', socket.id);
                if(socket.id === portais.id) {
                  // console.log('Desconectei o: ', socket.id);
                  socket.disconnect(true);
                }
              });
            }
          });
        }, 20000)
      });
      /** Garante que o portal está conectado */

      socket.on('login', (data) => {
        console.log('Login: ');
        this.login(socket, data);
      });

      /**
       * Exportação de vendas
       */
      socket.on('iniciaExportacao', (data) => {
        try {
          data.payload = JSON.parse(data.payload);
          UpService.setInsereServidor(data.CDFILIAL, data.CDCAIXA, data.CDLOJA, data.payload);
        } catch (e) {}
      });

      /**
       * Importação
       */
      socket.on('importar', (data) => {
        console.log('importar', data);
        let target = this.slavelist[socket.email][data.CDFILIAL];
        this.io
          .to(target.id)
          .emit('command', { type: 'import', data: { dataUrl: '/API/getStructure' } });
      });

      /**
       * Desconet do socket
       */
      socket.on('disconnect', () => {
        /** Limpa os Intervals */
        this.intervals.forEach((interval) => {
          if(interval.id === socket.id) {
            clearInterval(interval.interval);
          }
        });
        const indIntervals = this.intervals.findIndex((item) => {
          return item.id === socket.id;
        });
        this.intervals.splice(indIntervals);

        /** Limpa as Conexões (Portais e Lojas) */
        const indConexoes = this.conexoes.findIndex((item) => {
          return item.id === socket.id;
        });
        this.conexoes.splice(indConexoes);

        const subsidiaryIndex = this.slavelist.findIndex((item) => {
          return item.socketId === socket.id;
        });

        if (this.slavelist[subsidiaryIndex]) {
          this.slavelist[subsidiaryIndex].situacao = 'Inativa';

          const portalList = Object.keys(this.clientlist).map((item) => {
            return this.clientlist[item].id;
          });
          this.io.to(portalList).emit('login', {
            text: 'UNIDADE DESCONECTADA',
            error: false,
            message: {
              data: this.slavelist,
            },
          });
        }
      });

      //Refaz os intervalos de monitoramento para sempre buscar atualizações que possam ser feitas na tela de Parametrizações.
      this.assIntervals.forEach((interval) => {
        if(interval.id === socket.id || interval.id === this.lastSocketId) {
          clearInterval(interval.interval);
        }
      });

      this.getAssConfig().then((assConfig) => {
        assConfig.forEach((config) => {
          socket.on(config.ID, (data) => {
            this.assIntervals.push({
              id: socket.id,
              interval: setInterval(() => {
                monitoringAssistant.bind(this)(config);
              },
              intervalMonitoring(data))
            });
          });
          
          socket.on(`${config.ID}_detail`, (data) => {
            console.log('Filial:', data.CDFILIAL, '| Portal Monitoramento Detalhe:', config.ID + ' - ' + config.NOME);
            this.io.to(socket.id).emit(`${config.ID}_detail`, this.monitoringResult[data.CDFILIAL][config.ID]);
          });
        });
      });

      function monitoringAssistant(config) {
        console.log('Portal Monitoramento:', config.ID + ' - ' + config.NOME);
        this.io.to(socket.id).emit(config.ID, this.slavelist);
      }

      /**
       * Socket Portal
       * Socket Update
       */
      socket.on('updater', (data) => {
        console.log('Portal Update: ', new Date().getSeconds());
        /**
         * Executa quando abri o Grid
         */
        updaterController.getUnidades().then((result) => {
          console.log('Portal Update 1: ', new Date().getSeconds());

          setInterval(() => {
              this.io.to(socket.id).emit('updater', {
                text: 'UPDATE',
                error: false,
                message: {
                  data: result,
                },
              });
            }, 5000);

          //setInterval(() => {
            // let newItens = [];
            // result.forEach((results) => {
            //   let newR = results;
            //   let atu = 'Atualização pendente para a versão ';
            //   let atu2 = results.UPSISVERDISPONIVEL;
            //   updaterController.getVersaoAtual(results.IDUPDATERSISTEMAS).then((resultss) => {
            //     if(resultss.UPVERSTATUS === 0){
            //       results.UPSISVERDISPONIVEL = atu + atu2;
            //     }
            //     results.VERSAOATUAL = resultss;
            //     newItens.push(results);
            //     console.log('Portal Update 2: ', new Date().getSeconds(), newItens);
            //   });
            // });

            // setTimeout(() => {
            //   this.io.to(socket.id).emit('updater', {
            //     text: 'UPDATE',
            //     error: false,
            //     message: {
            //       data: newItens,
            //     },
            //   });
            // }, 3000);
          //}, intervalMonitoring({interval:5000}));
        }).catch((err) => {
          console.log(err);
        });
      });

      /**
       * Histórico de Atualizações
       */
      socket.on('updaterHistorico', (data) => {
        updaterController.getHistorico(data).then((result) => {
          console.log(result);
          this.io.to(socket.id).emit('updaterHistorico', {
            text: 'UPDATE',
            error: false,
            message: {
              data: result,
            },
          });
        });
      });

      /**
       * Ações do Update
       */
      socket.on('updaters', (data) => {
        console.log('updaters: ', data);
        if (data.tipo === 'up') {
          updaterController.updaterSistemas(data).then((result) => {
            socket.emit('updaters', {
              text: 'UPDATE',
              error: false,
              message: {
                data: result,
              },
            });
            console.log('result', result);
          }).catch((err) => {
            console.log(err);
          });
        } else if (data.tipo === 'down') {
          updaterController.getUnidades().then((result) => {
            console.log(result);
          }).catch((err) => {
            console.log(err);
          });
        } else if (data.tipo === 'historico') {
          updaterController.getUnidades().then((result) => {
            console.log(result);
          }).catch((err) => {
            console.log(err);
          });
        } else if (data.tipo === 'install') {
          updaterController.getUnidades().then((result) => {
            console.log(result);
          }).catch((err) => {
            console.log(err);
          });
        }
      });

      this.lastSocketId = socket.id;
    });

    /** Retorna Array com todos os socketId e Salas das lojas */
    function idsLojas(clientlist) {
      let ids = [];
      Object.keys(clientlist).map((item) => {
        if(clientlist[item].socketId){
          /** Por iD */
          ids.push(clientlist[item].socketId);
          /** Por Sala */
          ids.push(`${clientlist[item].nrorg}_${clientlist[item].id}`);
        }
      });
      return ids;
    }
  }

  import(socket, data) {}

  login(socket, data) {
    // Não é uma filial que está conectando e sim um portal
    if (!data.CDFILIAL) {
      // TODO Autenticar esse cliente, senão desconectar ele

      let auth = AuthService.autenticar(data.email, data.password)
        .then((result) => {
          // Se estiver autenticado coloca ele na lista simples de controle
          socket.email = data.email;

          this.clientlist[data.email] = socket;
          this.io
            .to(socket.id)
            .emit('confirmLogin', { error: false, message: 'Você está conectado', result: result });

          if (!data.slave) {
            this.conexoes.push({
              id: socket.id,
              time: new Date().getTime(),
            });
            if (this.slavelist === undefined) {
              this.getDashboardFiliais(data).then((dashboard) => {
                this.io.to(socket.id).emit('login', {
                  text: 'PORTAL CONECTADO',
                  error: false,
                  message: dashboard,
                  result,
                });
              });
            } else {
              this.io.to(socket.id).emit('login', {
                text: 'PORTAL CONECTADO',
                error: false,
                message: this.slavelist,
                result,
              });
            }
          }
        })
        .catch((error) => {
          this.io.to(socket.id).emit('message', 'Você não é um usuário autorizado!');
          socket.disconnect();
        });
    } else {
      // Uma filial se coneta ao master
      // TODO Autenticar esse cliente, senão desconectar ele
      let auth = AuthService.autenticar(data.email, '', data.CDFILIAL)
        .then((result) => {
          // Se estiver autenticado coloca ele na lista simples de controle
          socket.email = data.email;
          socket.CDFILIAL = data.CDFILIAL;

          if(!this.monitoringResult[socket.CDFILIAL]) {
            this.monitoringResult[socket.CDFILIAL] = {};
          }

          const subsidiaryIndex = this.slavelist.findIndex((item) => {
            return item.codigo === data.CDFILIAL;
          });

          if (this.slavelist[subsidiaryIndex]) {
            this.slavelist[subsidiaryIndex].situacao = 'Ativo';
            this.slavelist[subsidiaryIndex].socketId = socket.id;

            /** Registra a conexão */
            this.conexoes.push({
              id: socket.id,
              time: new Date().getTime(),
            });

            const portalList = Object.keys(this.clientlist).map((item) => {
              return this.clientlist[item].id;
            });

            /**
             * Socket Loja
             */
            this.io.to(socket.id).emit('message', result);
            this.io
              .to(socket.id)
              .emit('command', { type: 'downloadFromUrl', data: { dataUrl: '/API/getStructure' } });

            if (portalList.length) {
              /** Socket Loja */
              this.io.to(portalList).emit('login', {
                text: 'UNIDADE CONECTADA',
                error: false,
                message: {
                  data: this.slavelist,
                },
                result,
              });
            }

            // Executar a primeira vez mais rápido assim que conectar o socket
            setTimeout(() => {
              monitoringAssistant.bind(this)();
            }, this.defaultInitTimeout);

            // Cria os intervalos de monitoramento
            this.intervals.push({
              id: socket.id,
              interval: setInterval(() => {
                monitoringAssistant.bind(this)();
              },
              intervalMonitoring())
            });
      
            async function monitoringAssistant() {
              try {
                const assConfig = await this.getAssConfig();
                assConfig.forEach((config) => {                
                  // A => Execução na base do Assistant
                  // L => Execução na base Lojas, envia a requisição e aguarda o retorno para emitir para o Assistant
                  // I => Execução de rotinas criadas internamente que não podem ser alteradas via sistema (SAT, por exemplo)
                  // C => Execução de requisição HTTP nas Lojas, envia a requisição e aguarda o retorno para emitir para o Assistant
                  // E => Execução de requisição HTTP chamada pelo Assistant
                  // S => Verificação se Serviços do Windows na Loja estão em execução
                  if(config.TIPO == 'A') {
                    monitoringDataBaseCloud.bind(this)(config, subsidiaryIndex);
                  } else if(config.TIPO == 'L') {
                    monitoringDataBaseLoja.bind(this)(config, subsidiaryIndex);
                  } else if(config.TIPO == 'I') {
                    monitoringAssistantInterno.bind(this)(config, subsidiaryIndex);
                  } else if(config.TIPO == 'C') {
                    monitoringApacheCloud.bind(this)(config, subsidiaryIndex);
                  } else if(config.TIPO == 'E') {
                    monitoringApacheLoja.bind(this)(config, subsidiaryIndex);
                  } else if(config.TIPO == 'S') {
                    monitoringServicesLoja.bind(this)(config, subsidiaryIndex);
                  }
                });
              } catch (error) {
                console.log('ERRO NA CRIAÇÃO DO MONITORAMENTO', error);
              };
            };

            function monitoringDataBaseCloud(config, filialIdx) {
              const params = { CDFILIAL: socket.CDFILIAL };
              monitoringController.execCommand(config.COMANDO, params).then((response) => {
                console.log('Filial:', socket.CDFILIAL, '| DB Cloud SUC:', config.ID + ' - ' + config.NOME, '| Quantidade:', response.rows.length);
                setMonitoringResult.bind(this)(config, filialIdx, response, false);
              }).catch((error) => {
                const response = {
                  metaData: [ { name: 'ERRO' } ],
                  rows: [ [ error.message ] ]
                };
                console.log('Filial:', socket.CDFILIAL, '| DB Cloud ERR:', config.ID + ' - ' + config.NOME, '| Message:', error.message);
                setMonitoringResult.bind(this)(config, filialIdx, response, true);
              });
            }

            function monitoringApacheCloud(config, filialIdx) {
              monitoringController.execHttpRequest(config.COMANDO).then((response) => {
                console.log('Filial:', socket.CDFILIAL, '| HTTP Cloud SUC:', config.ID + ' - ' + config.NOME);
                setMonitoringResult.bind(this)(config, filialIdx, response, false);
              }).catch((error) => {
                const response = {
                  metaData: [ { name: 'ERRO' } ],
                  rows: [ [ `${error.message} => URL: ${config.COMANDO}` ] ]
                };
                console.log('Filial:', socket.CDFILIAL, '| HTTP Cloud ERR:', config.ID + ' - ' + config.NOME, '| Message:', `${error.message} => URL: ${config.COMANDO}`);
                setMonitoringResult.bind(this)(config, filialIdx, response, true);
              });
            }

            function monitoringDataBaseLoja(config, filialIdx) {
              //Envia para a Loja
              socket.emit('execMonitoring', config, socket.rooms);

              //Recebe a resposta da Loja, sempre atualizando o listener para não duplicar
              socket.removeAllListeners(`${config.ID}_result`);
              socket.on(`${config.ID}_result`, (result) => {
                if(!result.isError) {
                  console.log('Filial:', socket.CDFILIAL, '| SQL Loja SUC:', config.ID + ' - ' + config.NOME, '| Quantidade:', result.response.rows.length);
                  setMonitoringResult.bind(this)(config, filialIdx, result.response, false);
                } else {
                  const response = {
                    metaData: [ { name: 'ERRO' } ],
                    rows: [ [ result.response ] ]
                  };
                  console.log('Filial:', socket.CDFILIAL, '| SQL Loja ERR:', config.ID + ' - ' + config.NOME, '| Message:', result.response);
                  setMonitoringResult.bind(this)(config, filialIdx, response, true);
                }
              });
            }

            function monitoringApacheLoja(config, filialIdx) {
              //Envia para a Loja
              socket.emit('execMonitoringHttp', config, socket.rooms);

              //Recebe a resposta da Loja, sempre atualizando o listener para não duplicar
              socket.removeAllListeners(`${config.ID}_result`);
              socket.on(`${config.ID}_result`, (result) => {
                if(!result.isError) {
                  console.log('Filial:', socket.CDFILIAL, '| HTTP Loja SUC:', config.ID + ' - ' + config.NOME);
                  setMonitoringResult.bind(this)(config, filialIdx, result.response, false);
                } else {
                  const response = {
                    metaData: [ { name: 'ERRO' } ],
                    rows: [ [ result.response ] ]
                  };
                  console.log('Filial:', socket.CDFILIAL, '| HTTP Loja ERR:', config.ID + ' - ' + config.NOME, '| Message:', result.response);
                  setMonitoringResult.bind(this)(config, filialIdx, response, true);
                }
              });
            }

            function monitoringServicesLoja(config, filialIdx) {
              //Envia para a Loja
              socket.emit('execMonitoringServices', config, socket.rooms);

              //Recebe a resposta da Loja, sempre atualizando o listener para não duplicar
              socket.removeAllListeners(`${config.ID}_result`);
              socket.on(`${config.ID}_result`, (result) => {
                if(!result.isError) {
                  const isServicesError = result.response.rows.some((element) => {
                    return element.STATUS == 'N';
                  });
                  if(!isServicesError) {
                    result.response.rows = [];
                  }
                  console.log('Filial:', socket.CDFILIAL, '| Services Loja SUC:', config.ID + ' - ' + config.NOME);
                  setMonitoringResult.bind(this)(config, filialIdx, result.response, isServicesError);
                } else {
                  const response = {
                    metaData: [ { name: 'ERRO' } ],
                    rows: [ [ result.response ] ]
                  };
                  console.log('Filial:', socket.CDFILIAL, '| Services Loja ERR:', config.ID + ' - ' + config.NOME, '| Message:', result.response);
                  setMonitoringResult.bind(this)(config, filialIdx, response, true);
                }
              });
            }

            function setMonitoringResult(config, filialIdx, response, isError) {
              //Atualiza a lista a ser usada no grid de Monitoramento.
              if(this.slavelist[filialIdx]) {
                this.slavelist[filialIdx][config.ID] = isError || response.rows.length > 0 ? 'N' : 'S';
              }
              //Atualiza os dados do modal de detalhes de cada monitoramento.
              this.monitoringResult[socket.CDFILIAL][config.ID] = response;
            }

            function monitoringAssistantInterno(config, filialIdx) {
              if(config.NOME == 'SAT' && this.slavelist[filialIdx].sat !== 'N') {
                //Envia para a Loja
                socket.emit("satMonitoring", data, socket.rooms);

                //Recebe a resposta da Loja, sempre atualizando o listener para não duplicar
                socket.removeAllListeners('satMonitoringResult');
                socket.on('satMonitoringResult', (result) => {
                  if(!result.isError) {
                    const isSatError = result.response.rows.some((element) => {
                      return !element.sat;
                    });
                    if(!isSatError) {
                      result.response.rows = [];
                    }
                    console.log('Filial:', socket.CDFILIAL, '| Exec SAT SUC:', config.ID + ' - ' + config.NOME, '| Error:', isSatError);
                    setMonitoringResult.bind(this)(config, filialIdx, result.response, isSatError);
                    if(this.slavelist[filialIdx]) {
                      this.slavelist[filialIdx].sat = isSatError ? 'E' : 'S';
                    }
                  } else {
                    const response = {
                      metaData: [ { name: 'ERRO' } ],
                      rows: [ [ result.response ] ]
                    };
                    console.log('Filial:', socket.CDFILIAL, '| Exec SAT ERR:', config.ID + ' - ' + config.NOME, '| Message:', result.response);
                    setMonitoringResult.bind(this)(config, filialIdx, response, true);
                    if(this.slavelist[filialIdx]) {
                      this.slavelist[filialIdx].sat = 'E';
                    }
                  }
                });
              }
            }
          }
        })
        .catch((error) => {
          this.io.to(socket.id).emit('message', 'Você não é um usuário autorizado!');
          socket.disconnect();
        });
    }
  }

  async getDashboardFiliais(data = null) {
    const filiais = [];
    const queryResultAll = await adqGetDashboardFiliais.get(paginacao(data));
    const { rows } = queryResultAll;
    const assConfig = await this.getAssConfig();

    rows.forEach((element) => {
      const filialRow = {
        id: element[0],
        codigo: element[0],
        name: element[1],
        situacao: 'Inativa',
        email: '',
        socketId: '',
        sat: element[3] == 0 ? 'N' : 'E', // Tratamento especial para o Sat: S => Sucesso | E => Error | N => Filial não utiliza SAT
        websocket: '',
        nrorg: element[2],
      };

      //Cria todos os monitoramentos dinâmicos na row para funcionamento do filtro | N => Erro (bolinha vermelha)
      assConfig.forEach((config) => {
        filialRow[config.ID] = 'N'
      });
      filiais.push(filialRow);
    });

    if (this.slavelist !== undefined) { filiais.concat(this.slavelist); }

    return filiais;
  }

  async getAssConfig() {
    const config = [];
    const assConfig = await parametrizacoesModel.getAssConfig(true);
    assConfig.rows.forEach((element) => {
      config.push({
        ID:        element[0].toString(),
        NOME:      element[1],
        DESCRICAO: element[2],
        TIPO:      element[3],
        ORDEM:     element[4],
        ATIVO:     element[5],
        COMANDO:   element[6],
        INTERVALO: element[7]
      });
    });
    return config;
  }
}

module.exports = new SocketService();
